/*
 * Copyright 2008, Systemic Realtime Design LLC. All Rights Reserved.
 *
 * This source code may contain confidential information of Systemic
 * Realtime Design LLC and its licensors. Any use, reproduction,
 * modification, disclosure, distribution or transfer of this software,
 * or any software which includes or is based upon any of this code, is
 * prohibited unless expressly authorized by SRD by written agreement. For
 * more information (including whether this source code file has been
 * published) please email licensing@sysrtime.com.
 */

/*
 * net/tipc/tipc_core.h: Include file for TIPC global declarations
 * 
 * Copyright (c) 2005-2006, Ericsson AB
 * Copyright (c) 2005-2007, Wind River Systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the names of the copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _TIPC_CORE_H
#define _TIPC_CORE_H

#define _KERNEL 1
#define KERNEL  1
#define __KERNEL__ 1

#include <sys/types.h>
#include <sys/neutrino.h>
#include <sync.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
/* fix compilation problem on qnx 6.4.0 */
//#include <stdlib.h>
#include <sys/times.h>

/* fix compilation problem on qnx 6.4.0 */
#undef bcopy
#undef bzero
#undef bcmp

/* MLL@SRD: need to define M_WAITOK or else mbuf.h brings in sys/malloc.h which 
 *          has different kernel? versions of malloc with 3 args
 */
#define M_WAITOK	0x0000
#define M_NOWAIT    0x0001
#define M_ZERO      0x0002
#define M_CANFAIL   0x0004


#include <linux/tipc.h>
#include <linux/tipc_config.h>
#include <net/tipc/tipc_plugin_msg.h>
#include <net/tipc/tipc_plugin_port.h>
#include <net/tipc/tipc_plugin_if.h>
#include <net/tipc/tipc.h>


#define TIPC_MOD_VER "1.7.5"

#define CONFIG_TIPC_UNICLUSTER_FRIENDLY
#define CONFIG_TIPC_MULTIPLE_LINKS
#define CONFIG_TIPC_DEBUG
#define CONFIG_TIPC_SYSTEM_MSGS
#define CONFIG_TIPC_CONFIG_SERVICE
#define CONFIG_TIPC_SOCKET_API


/*
 * TIPC emulation of common file atomic operations API
 */

#define ATOMIC_INIT(i) (i)
typedef int atomic_t;

#define atomic_read(pX) (*pX)

#define atomic_set(pX, val) \
    { InterruptDisable (); (*pX) = val; InterruptEnable(); }

#define atomic_inc(pX) \
    { InterruptDisable (); (*pX)++; InterruptEnable(); }

#define atomic_dec(pX) \
    { InterruptDisable (); (*pX)--; InterruptEnable(); }

static inline atomic_t atomic_clear (atomic_t *pX)
{
	InterruptDisable();
	int res = *pX;
	*pX = 0;
	InterruptEnable();
	return res;
}

static inline int atomic_inc_return (atomic_t *pX)
{
	int retval;
	InterruptDisable();
	retval = ++(*pX);
	InterruptEnable();
	return retval;
}


/*
 * TIPC emulation of common file memory operations API
 */
#define GFP_ATOMIC 0

#define kmalloc(size, dummy)	((void *)malloc (size))
#define kfree(ptr)              (free (ptr))
#define vmalloc(size)           (kmalloc (size, 0))
#define vfree(ptr)              (kfree (ptr))
static inline void *kzalloc(int size, int dummy)
{
	void *p;
	p = calloc(1, size);
	if (p) {
		memset(p, 0, size);
	}
	return p;
}

static inline void *kcalloc(int n, int size, int dummy)
{
	return kzalloc (n*size, dummy);
}

static inline unsigned long copy_from_user(void *to, const void *from, unsigned long n)
{
    /* No need to validate address range, since system call code handles this */
    memcpy (to, from, n);
    return 0;
}



/*
 * Spinlock wrappers (lets TIPC common files run unchanged on other OS's)
 */
 
#define DECLARE_SPINLOCK(x) extern pthread_mutex_t x
#define DECLARE_RWLOCK(x)   extern pthread_rwlock_t x

#define DEFINE_SPINLOCK(x) pthread_mutex_t  x = PTHREAD_MUTEX_INITIALIZER;
#define DEFINE_RWLOCK(x)   pthread_rwlock_t x = PTHREAD_RWLOCK_INITIALIZER;

#define write_lock_bh(lockp)     pthread_rwlock_wrlock(lockp)
#define write_unlock_bh(lockp)   pthread_rwlock_unlock(lockp)
#define read_lock_bh(lockp)      pthread_rwlock_rdlock(lockp)
#define read_unlock_bh(lockp)    pthread_rwlock_unlock(lockp)

#define spin_lock_init(lockp)    pthread_mutex_init(lockp, NULL)
#define spin_lock_bh(lockp)      pthread_mutex_lock(lockp)
#define spin_unlock_bh(lockp)    pthread_mutex_unlock(lockp)
#define spin_trylock_bh(lockp)   pthread_mutex_trylock(lockp)
#define spin_lock_term(lockp)    pthread_mutex_destroy(lockp)
#define spin_lock_destroy(lockp) pthread_mutex_destroy(lockp)


/*
 * Sanity test macros
 */

#ifdef CONFIG_TIPC_DEBUG
#define dbg_assert(i) assert(i)
#else
#define dbg_assert(i) assert(i)
#endif

/*
 * TIPC system monitoring code
 */

/*	
 * TIPC's print buffer module supports the following print buffers:
 *
 * TIPC_NULL : null buffer (i.e. print nowhere)
 * TIPC_CONS : system console
 * TIPC_LOG  : TIPC log buffer
 * &buf	     : user-defined buffer (struct print_buf *)
 *
 * Note: TIPC_LOG is configured to echo its output to the system console;
 *       user-defined buffers can be configured to do the same thing.
 */

extern struct print_buf *const TIPC_NULL;
extern struct print_buf *const TIPC_CONS;
extern struct print_buf *const TIPC_LOG;

void tipc_printf(struct print_buf *, const char *fmt, ...);

/*
 * TIPC_OUTPUT is the destination print buffer for system messages.
 */

#ifndef TIPC_OUTPUT
#define TIPC_OUTPUT TIPC_LOG
#endif			  

/*
 * TIPC can be configured to send system messages to TIPC_OUTPUT, the system
 * console only, or to suppress them entirely.
 */

#ifdef CONFIG_TIPC_SYSTEM_MSGS

#ifdef CONFIG_TIPC_CONFIG_SERVICE

#define printk printf

#define err(fmt, arg...)  tipc_printf(TIPC_OUTPUT, "TIPC: " fmt, ## arg)
#define warn(fmt, arg...) tipc_printf(TIPC_OUTPUT, "TIPC: " fmt, ## arg)
#define info(fmt, arg...) tipc_printf(TIPC_OUTPUT, "TIPC: " fmt, ## arg)

#else

#define err(fmt, arg...)  printk("TIPC: " fmt , ## arg)
#define info(fmt, arg...) printk("TIPC: " fmt , ## arg)
#define warn(fmt, arg...) printk("TIPC: " fmt , ## arg)

#endif

#else

#define err(fmt, arg...)  do {} while (0)
#define info(fmt, arg...) do {} while (0)
#define warn(fmt, arg...) do {} while (0)

#endif

/*
 * DBG_OUTPUT is the destination print buffer for debug messages.
 * It defaults to the the null print buffer, but can be redefined
 * (typically in the individual .c files being debugged) to allow
 * selected debug messages to be generated where needed.
 */

#ifndef DBG_OUTPUT
#define DBG_OUTPUT TIPC_NULL
#endif

/*
 * TIPC can be configured to send debug messages to the specified print buffer
 * (typically DBG_OUTPUT) or to suppress them entirely.
 */

#ifdef CONFIG_TIPC_DEBUG

#define dbg_printf(pb, fmt, arg...) \
	do {if (pb != TIPC_NULL) tipc_printf(pb, fmt, ## arg);} while (0)
#define dbg_msg(pb, msg, txt) \
	do {if (pb != TIPC_NULL) tipc_msg_dbg(pb, msg, txt);} while (0)
#define dbg_dump(pb, fmt, arg...) \
	do {if (pb != TIPC_NULL) tipc_dump_dbg(pb, fmt, ##arg);} while (0)

#define dbg(fmt, arg...)	dbg_printf(DBG_OUTPUT, fmt, ##arg)
#define msg_dbg(msg, txt)	dbg_msg(DBG_OUTPUT, msg, txt)
#define dump_dbg(fmt, arg...)	dbg_dump(DBG_OUTPUT, fmt, ##arg)

void tipc_msg_dbg(struct print_buf *, struct tipc_msg *, const char *);
void tipc_dump_dbg(struct print_buf *, const char *fmt, ...);

#else

#define dbg_printf(pb, fmt, arg...)	do {} while (0)
#define dbg_msg(pb, msg, txt)		do {} while (0)
#define dbg_dump(pb, fmt, arg...)	do {} while (0)

#define dbg(fmt, arg...)	do {} while (0)
#define msg_dbg(msg, txt)	do {} while (0)
#define dump_dbg(fmt, arg...)	do {} while (0)

#define tipc_msg_dbg(...)	do {} while (0)
#define tipc_dump_dbg(...)	do {} while (0)

#endif


/* 
 * TIPC-specific error codes
 */

#define ELINKCONG EAGAIN	/* link congestion <=> resource unavailable */

/*
 * Global configuration variables
 */

extern u32 tipc_own_addr;
extern int tipc_max_nodes;
extern int tipc_max_clusters;
extern int tipc_max_zones;
extern int tipc_max_remotes;
extern int tipc_max_ports;
extern int tipc_max_subscriptions;
extern int tipc_max_publications;
extern int tipc_net_id;
extern int tipc_remote_management;

/*
 * Other global variables
 */

extern int tipc_mode;
extern int tipc_random;
extern const char tipc_alphabet[];
extern atomic_t tipc_user_count;


/*
 * Routines available to privileged subsystems
 */

extern int  tipc_core_start(void);
extern void tipc_core_stop(void);
extern int  tipc_core_start_net(void);
extern void tipc_core_stop_net(void);


static inline int delimit(int val, int min, int max)
{
	if (val > max)
		return max;
	if (val < min)
		return min;
	return val;
}




/*
 * TIPC emulation of common file linked list API
 */

#define container_of(ptr, type, member) \
    ((type *)((char *)(ptr) - offsetof(type,member)))

struct list_head {
    struct list_head *next;
    struct list_head *prev;
};

#define LIST_HEAD_INIT(name) { &(name), &(name) }

/* MLL@SRD: "LIST_HEAD" has namespace conflict with "sys/proc.h" */
#define TIPC_LIST_HEAD(name) \
    struct list_head name = LIST_HEAD_INIT(name)

#define INIT_LIST_HEAD(ptr) do { \
        (ptr)->next = (ptr); (ptr)->prev = (ptr); \
    } while (0)

static inline int list_empty(struct list_head *head)
{
    return head->next == head;
}

static inline void __list_add(struct list_head *new,
			      struct list_head *prev,
			      struct list_head *next)
{
    next->prev = new;
    new->next = next;
    new->prev = prev;
    prev->next = new;
}

static inline void list_add(struct list_head *new, struct list_head *head)
{
    __list_add(new, head, head->next);
}

static inline void list_add_tail(struct list_head *new, struct list_head *head)
{
    __list_add(new, head->prev, head);
}

static inline void __list_del(struct list_head *prev, struct list_head *next)
{
    next->prev = prev;
    prev->next = next;
}

static inline void list_del(struct list_head *entry)
{
    __list_del(entry->prev, entry->next);
    entry->next = (void *) 0;
    entry->prev = (void *) 0;
}

static inline void list_del_init(struct list_head *entry)
{
    __list_del(entry->prev, entry->next);
    INIT_LIST_HEAD(entry); 
}

static inline void list_move_tail(struct list_head *list,
				  struct list_head *head)
{
    __list_del(list->prev, list->next);
    list_add_tail(list, head);
}

#define list_entry(ptr, type, member) container_of(ptr, type, member)

#define list_for_each_entry(pos, head, member) \
	for (pos = list_entry((head)->next, typeof(*pos), member); \
	     &pos->member != (head); \
	     pos = list_entry(pos->member.next, typeof(*pos), member))

#define list_for_each_entry_safe(pos, n, head, member) \
	for (pos = list_entry((head)->next, typeof(*pos), member), \
		n = list_entry(pos->member.next, typeof(*pos), member); \
	     &pos->member != (head); \
	     pos = n, \
		n = list_entry(n->member.next, typeof(*n), member))
        
#define list_for_each_safe(pos, n, head) \
    for (pos = (head)->next, n = pos->next; pos != (head); \
                 pos = n, n = pos->next)

struct hlist_node
{
    struct hlist_node *next;
    struct hlist_node **pprev;
};

struct hlist_head
{
    struct hlist_node *first;
};

static inline int hlist_empty(struct hlist_head *head)
{
    return (!head->first);
}

static inline void __list_splice(struct list_head *list,
                                 struct list_head *head)
{
    struct list_head *first = list->next;
    struct list_head *last = list->prev;
    struct list_head *at = head->next;

    first->prev = head;
    head->next = first;

    last->next = at;
    at->prev = last;
}

static inline void list_splice(struct list_head *list, struct list_head *head)
{
    if (!list_empty(list))
        __list_splice(list, head);
}

#define INIT_HLIST_NODE(ptr) ((ptr)->next = NULL, (ptr)->pprev = NULL)

#define hlist_entry(ptr, type, member) container_of(ptr, type, member)

static inline void __hlist_del(struct hlist_node *n) 
{
    struct hlist_node *next = n->next;
    struct hlist_node **pprev = n->pprev;
    *pprev = next;  
    if (next) 
        next->pprev = pprev;
}  

static inline void hlist_del_init(struct hlist_node *n) 
{
    if (n->pprev)
	{
        __hlist_del(n);
        INIT_HLIST_NODE(n);
        }
}  

static inline void hlist_add_head(struct hlist_node *n, struct hlist_head *h) 
{ 
    struct hlist_node *first = h->first;
    n->next = first; 
    if (first) 
        first->pprev = &n->next;
    h->first = n; 
    n->pprev = &h->first; 
} 

#define hlist_for_each_entry(tpos, pos, head, member) \
	for (pos = (head)->first; \
	     pos && (((tpos = hlist_entry(pos, typeof(*tpos), member)), 1)); \
	     pos = pos->next)

#define hlist_for_each_entry_safe(tpos, pos, n, head, member) \
	for (pos = (head)->first; \
	     pos && ((n = pos->next), 1) && \
		((tpos = hlist_entry(pos, typeof(*tpos), member)), 1); \
	     pos = n)





/*
 * TIPC timer and signal code
 */

#include "tipc_timer.h"

/*
 * TIPC socket buffer code
 */

#include "tipc_skbuff.h"


/**
 * buf_acquire - creates a TIPC message buffer
 * @size: message size (including TIPC header)
 *
 * Returns a new buffer with data pointers set to the specified size.
 * 
 * NOTE: Headroom is reserved to allow prepending of a data link header.
 *       There may also be unrequested tailroom present at the buffer's end.
 */

struct sk_buff *buf_acquire(u32 size);

/**
 * buf_discard - frees a TIPC message buffer
 * @skb: message buffer
 *
 * Frees a message buffer.  If passed NULL, just returns.
 */

static inline void buf_discard(struct sk_buff *skb)
{
	if (likely(skb != NULL))
		kfree_skb(skb);
}

/**
 * buf_linearize - convert a TIPC message buffer into a single contiguous piece
 * @skb: message buffer
 * 
 * Returns 0 on success.
 *
 * MLL@SRD: buffer is always linear, so operation cannot fail
 */

static inline int buf_linearize(struct sk_buff *skb)
{
	return 0;
}

#endif
