/*
 * Copyright 2008, Systemic Realtime Design LLC. All Rights Reserved.
 *
 * This source code may contain confidential information of Systemic
 * Realtime Design LLC and its licensors. Any use, reproduction,
 * modification, disclosure, distribution or transfer of this software,
 * or any software which includes or is based upon any of this code, is
 * prohibited unless expressly authorized by SRD by written agreement. For
 * more information (including whether this source code file has been
 * published) please email licensing@sysrtime.com.
 */

/* ------------------------------------------------------------------------
 *
 * tipc_sk_buff.c
 *
 * Short description: TIPC buffer management routines
 * 
 * ------------------------------------------------------------------------
 * Copyright (c) 2003, Ericsson Research Canada
 * Copyright (c) 2006-2007 Wind River Systems, Inc.
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this 
 * list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * Neither the names of the copyright holders nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ------------------------------------------------------------------------
 *
 *  Created 2003-11-15 by Jon Maloy
 *
 * ------------------------------------------------------------------------
 *
 *
 *  Rev	Date		Rev by			Reason
 *  ---	----		------			------
 *
 *  PA1	2003-11-15	Jon Maloy		Created.
 *
 * -----------------------------------------------------------------------
 * modification history
 * --------------------
 * 01e,25jul07,ebh  Change info to dbg for buffers allocated message
 * 01d,04apr07,mze  changes post SMP code review.
 * 01c,01dec06,mze  mods for TIPC 1.7.1 port.
 * 01b,15feb06,bwa  reworked skb_clone().
 * 01a,25jan06,ebh  Port of SF TIPC 1.6 code base, import buf.c changes
 * 
 */

/*
DESCRIPTION
Buffer management routines used by TIPC adaptation layer.  Each buffer is
used to hold a TIPC message, which consists of a TIPC header and the message
payload.  Both the TIPC header and the payload are of variable size.
The payload of a message may contain one or more TIPC messages that have
been bundled into a single message by TIPC; conversely, it may contain only
a portion of a larger message that has been divided into fragments.

INTERNAL
The TIPC core files require the use of a buffer structure called "sk_buff".
The sk_buff (socket buffer) structure defined in the Linux kernel is a complex 
structure that utilizes a single dynamically allocated byte array to hold 
the message.

The VxWorks version of TIPC utilizes a much simpler sk_buff structure, 
consisting of a pointer to an mBlk, a "next" pointer, a control block area, a
length field, a pointer to the start of data, and a pointer to the end of data.
To be consistent with the buffer paradigm assumed by the TIPC core files, 
the VxWorks implementation of sk_buff always uses a single mBlk tuple 
(consisting of an mBlk, clBlk, and cluster) for all message buffers;
this ensures that the message lies in a single contiguous byte array.

Wherever possible vxTIPC incorporates the mBlk tuple passed to it by the
driver code into an sk_buff.  (Such tuples are allocated from the driver pools.)
However, if vxTIPC receives a message that utilizes more than one tuple, it
converts the message into a single tuple message by allocating a tuple of the
appropriate size from the network stack data pool.  If the TIPC data pool does
not have an available cluster of the required size, vxTIPC will allocate the
cluster from the kernel heap area if possible.  This approach allows vxTIPC to
support TIPC messages of up to 66000 bytes even if the TIPC data pool is not
configured with clusters of that size.

To use storage efficiently, the sk_buffs used by vxTIPC are allocated from heap 
in blocks, and then distributed and reclaimed on an "as needed" basis.  
If the supply of sk_buffs becomes exhausted additional blocks of storage 
are allocated as required to increase the supply.  The first sk_buff in each 
block is used to chain the blocks together and cannot be used in a message 
buffer; the remaining sk_buffs in the block are added to the list of free 
sk_buffs and are available for use by applications.

The following statements are always true about a vxTIPC message buffer.
- The size of a message buffer is specified when the buffer is created,
  and limits how big a message may be placed in the buffer.
- The message contained in a message buffer need not utilize all of the
  available space in the buffer.  It is legal to have unused space
  at the end of the buffer (but not the beginning).
- The message buffer need not occupy all of the available space in the cluster;
  there may be unused space at the beginning and/or the end of the cluster.
  (Unused space at the beginning can sometimes be used to easily prepend 
  a message header to an existing message buffer.)
- If the size of a message buffer changes, this may involve a simple change
  to the data pointer and length fields of the associated mBlk.  However,
  in some cases it is necessary to allocate a new, larger cluster and copy 
  the existing data to it. 
- The size of the message buffer is given by the packet length field of
  the mBlk, which is equal to the mBlk length field.
- The mBlk for a message buffer always has the M_PKTHDR flag set 
  
*/

/* includes */

#define USE_VXWORKS_LISTS
/* fix compilation problem on qnx 6.4.0 */
//#include <stdlib.h>

#include "tipc_core.h"
#include "tipc_msg.h"
#include "tipc_dbg.h"
#include "tipc_skbuff.h"

#include "sys/param_bsd.h"
#include "sys/mbuf.h"

#include <sys/socket.h>
#include <sys/socketvar.h>

int numbufs=0;

/* defines */

#define MAX_SK_BUFFS    10000           /* max buffers allocated before cleanup */ 
#define NUM_SK_BUFFS    50              /* # sk_buffs per block */

/* external declarations */

/* local variables */

static struct sk_buff *sk_buffFirstBlock = NULL; /* first block of sk_buffs */
static struct sk_buff *sk_buffFirstFree = NULL;  /* first free sk_buff */

static int sk_buffCount = 0;             /* # of sk_buffs in existence */
static int sk_buffAlloc = 0;             /* # of sk_buffs currently in use */

static DEFINE_SPINLOCK(buffer_lock);


/*******************************************************************************
*
* buf_free - terminate buffer management subsystem
*
* This routine releases any buffer management subsystem data structures
* that were dynamically allocated.
*
*/
 
void buf_free (void)
{
    struct sk_buff *buf;                /* array of buffers */

    /* Release all blocks of buffers */
         
    spin_lock_bh (&buffer_lock);

    while (sk_buffFirstBlock != NULL)
    {
        buf = sk_buffFirstBlock;
        sk_buffFirstBlock = sk_buffFirstBlock->next;
        free (buf);
		buf = NULL;
    }
	sk_buffAlloc = 0;
    sk_buffCount = 0;
   
    spin_unlock_bh (&buffer_lock);
}

/*******************************************************************************
*
* vxskb_alloc_bufs - allocate a block of null buffers
*
* This routine allocates a block of sk_buffs from heap, and adds them to the
* pool of available sk_buffs.
*
* RETURNS: # of buffers allocated
*
*/

static int vxskb_alloc_bufs (void)
{
    struct sk_buff *buf;                /* newly allocated array of buffers */
    int count;                          /* loop counter */
    
    /* Allocate block of buffers */
    
    spin_unlock_bh (&buffer_lock);
    buf = (struct sk_buff *)malloc (NUM_SK_BUFFS * sizeof (struct sk_buff));
    spin_lock_bh (&buffer_lock);

    if (buf == NULL)
        {
        err ("Unable to allocate additional buffers\n");
        return 0;
        }

    /*
     * Add new buffers to pool of free buffers
     * (Note that the first buffer in the new block is used by 
     * the buffer management routines to chain blocks together)
     */
    
    buf->next = sk_buffFirstBlock;
    sk_buffFirstBlock = buf++;
    
    for (count = 1; count < NUM_SK_BUFFS; count++)
        {
        buf->mBufPtr = NULL;
        buf->next = sk_buffFirstFree;
        sk_buffFirstFree = buf++;
        }
    
    sk_buffCount += (NUM_SK_BUFFS - 1);
    
    PRINT_DBG("vxskb_alloc_bufs: %d buffers now allocated\n", sk_buffCount);
    
    return NUM_SK_BUFFS-1;
}

/*******************************************************************************
*
* vxskb_alloc - allocate a null buffer
*
* This routine allocates an sk_buff from the pool of available sk_buffs.
* The newly allocated sk_buff has no associated mBlk tuple.
*
* RETURNS: pointer to buffer, or NULL if no buffers available
*
*/

static struct sk_buff *vxskb_alloc (void)
{
    struct sk_buff *buf;                /* newly allocated buffer */

    spin_lock_bh (&buffer_lock);

    if ((sk_buffFirstFree == NULL) && (vxskb_alloc_bufs () == 0))
        {
        spin_unlock_bh (&buffer_lock);
        return NULL;
        }
    
    buf = sk_buffFirstFree;
    sk_buffFirstFree = buf->next;
    memset(buf, 0, sizeof (*buf));
    sk_buffAlloc++;

    spin_unlock_bh (&buffer_lock);
    return buf;
}

/*******************************************************************************
*
* vxskb_dealloc - deallocate a null buffer
*
* This routine returns "buf" to the pool of available sk_buffs.
* The buffer being freed must not have an associated mBlk tuple.
*
*/

void vxskb_dealloc
    (
    struct sk_buff *buf                 /* buffer to be freed */
    )
{
    spin_lock_bh (&buffer_lock);

    buf->mBufPtr = NULL;
    buf->next = sk_buffFirstFree;
    sk_buffFirstFree = buf;
    sk_buffAlloc--;

    spin_unlock_bh (&buffer_lock);
	
	/* cleanup if too many allocated bufs */
	if (sk_buffCount >= MAX_SK_BUFFS)
	{
		warn("too many bufs (%d), performing cleanup\n", sk_buffCount);
		buf_free();
	}
}

/*******************************************************************************
*
* vxskb_tuple_get - allocate a mBlk tuple for a message buffer 
*
* This routine allocates an mBlk tuple capable of holding "size" bytes
* of message content (i.e. header and payload combined).
*
* RETURNS: pointer to tuple, or NULL if unable to allocate one
*
*/

static mbuf_t* vxskb_tuple_get 
    (
    uint size                           /* size of cluster (in bytes) */
    )
{
    mbuf_t *mBufPtrNew;                 /* newly allocated mBlk */
 
    /* Allocate a tuple (must be able to hold a complete message header) */

    MGETHDR(mBufPtrNew, M_DONTWAIT, MT_DATA);
	if (!mBufPtrNew) {
		err("%s(): can't allocate mbuf\n", __FUNCTION__);
		goto failed;	
	}

    MCLGET(mBufPtrNew, M_DONTWAIT);
	if (!(mBufPtrNew->m_flags & M_EXT)) {
		err("%s(): can't allocate mcluster\n", __FUNCTION__);
		MBUF_FREE(mBufPtrNew);	
		goto failed;	
	}

    mBufPtrNew->m_flags |= M_PKTHDR;

    return mBufPtrNew;
    
failed:
    err ("Unable to allocate %d byte buffer\n", size);
    return NULL;
}


/*******************************************************************************
*
* vxskb_attach - attach mBlk chain to a new message buffer 
*
* This routine allocates a message buffer that incorporates the mBlk chain
* specified by "mBlkChain".  If the mBlk chain does not consist of a single
* mBlk tuple (as required by TIPC), it is converted to the required form.
*
* RETURNS: pointer to buffer, or NULL if unable to construct buffer
*
* INTERNAL
* This routine is very time critical.
*
*/

struct sk_buff *vxskb_attach 
    (
    mbuf_t *mBlkChain                  /* mBlk chain to incorporate */
    )
{
    struct sk_buff *buf;                /* newly allocated buffer */
    mbuf_t *mBlkCurr;                   /* current mBlk in chain */
    unchar *pData;                      /* ptr to 1st unused byte in buffer */
  	uint32_t len; 
   
  	/* get size */
	len = (mBlkChain->m_flags & M_PKTHDR)? mBlkChain->m_pkthdr.len : mBlkChain->m_len; 
    
    /* Allocate a buffer */
    
    buf = vxskb_alloc ();
    if (buf == NULL)
        return NULL;

    /* Attach mBlk chain if it consists of a single tuple, then exit */
    
    /* XXX it is quite possible that .mNext is always NULL, and the following
     * code block might not be needed
     */
    if (mBlkChain->m_next == NULL)
    {
        buf->len = len;
        buf->mBufPtr = mBlkChain;
        buf->data = (unsigned char *)buf->mBufPtr->m_data;
        buf->tail = buf->data + buf->len;
        return buf;
    }
  
    /* Allocate a new mBlk tuple of the required size */
    
    buf->mBufPtr = MBUF_DUP(mBlkChain, 0, M_COPYALL, M_DONTWAIT);
    if (buf->mBufPtr == NULL)
    {
        vxskb_dealloc (buf);
        return NULL;
    }
    buf->len = len;
    if(buf->mBufPtr->m_flags)
        buf->data = (unsigned char *)buf->mBufPtr->m_ext.ext_buf;
    else
        buf->data = NULL;
    if((buf->tail = buf->data))
        buf->tail += buf->len;

    MBUF_FREE (mBlkChain);

    return buf;
}


struct sk_buff *skb_copy
    (
    const struct sk_buff *skb,	/* sk_buff to copy */
    int gfp_mask		/* unused */
    )
{
    struct sk_buff *new_skb;
    int headerlen;

    new_skb = vxskb_alloc(); 
    if (!new_skb)
		return NULL;
    new_skb->next = NULL;
    new_skb->mBufPtr = MBUF_DUP(skb->mBufPtr, 0, M_COPYALL, M_DONTWAIT);
    if(!new_skb->mBufPtr)
    {
        vxskb_dealloc(new_skb);
        return NULL;
    }
    new_skb->data = new_skb->mBufPtr->m_data;
    headerlen = skb->data - vxskb_head (skb);
    skb_reserve (new_skb, headerlen);
    skb_put (new_skb, skb->len);
    
    /* note: doesn't copy contents of headroom area */
    memcpy (new_skb->cb, skb->cb, sizeof (skb->cb));

    return new_skb;
}

struct sk_buff *skb_clone
    (
    struct sk_buff *skb,	/* buffer to clone */
    int gfp_mask		/* unused */
    )
{
    struct sk_buff *clone;

    if (NULL == (clone = vxskb_alloc ()))
    {
		return NULL;
	}

    *clone = *skb;
    clone->next = NULL;
	clone->mBufPtr = MBUF_DUP (skb->mBufPtr, 0, M_COPYALL, M_DONTWAIT);	
    if (NULL == clone->mBufPtr)
    {
        vxskb_dealloc (clone);
        return NULL;
    }
    clone->data = clone->mBufPtr->m_data;
	return clone;
}


int skb_cow
    (
    struct sk_buff *skb,	/* buffer to copy-on-write */
    unsigned int headroom	/* needed headroom */
    )
{
    int curr_headroom = skb_headroom (skb);
    int needed = headroom - curr_headroom;

    if (needed < 0)
        needed = 0;

    if (needed || skb_cloned (skb))
    {
	    mbuf_t *new_mblk;
        int new_size;
        unchar *data = NULL;
	    needed = (needed + 3) & ~3;   /* round up to a multiple of 4 bytes */
        new_size = needed + vxskb_clsize (skb);
        new_mblk = vxskb_tuple_get (new_size);
        if (!new_mblk)
            return -ENOMEM;
        data = kmalloc(skb->len, GFP_ATOMIC);
        if(unlikely(!data))
            return -ENOMEM;
        m_copydata(skb->mBufPtr, 0, skb->len, data);
	    new_mblk->m_data += (needed + curr_headroom);
        m_copyback(new_mblk, 0, skb->len, data);
        MBUF_FREE(skb->mBufPtr);
        kfree(data);
	    skb->data = (unsigned char *)new_mblk->m_data;
	    skb->tail = skb->data + skb->len;
        skb->mBufPtr = new_mblk;
	    skb->mBufPtr->m_pkthdr.len = skb->len;
    }

    return 0;
}

void skb_reserve (struct sk_buff *skb, unsigned int len)
{
    skb->mBufPtr->m_data += len;
    skb->data += len;

    skb->tail += len;
}

unsigned char *skb_put (struct sk_buff *skb, unsigned int len)
{
    unsigned char *orig_tail = skb->tail;

    skb->mBufPtr->m_pkthdr.len += len;
    skb->mBufPtr->m_len += len;
    skb->len += len;

    skb->tail += len;
    return orig_tail;
}

unsigned char *skb_push (struct sk_buff *skb, unsigned int len)
{
    skb->mBufPtr->m_data -= len;
    skb->data -= len;

    skb->mBufPtr->m_pkthdr.len += len;
    skb->mBufPtr->m_len += len;
    skb->len += len;

    assert (skb->data >= vxskb_head (skb));
    return skb->data;
}

unsigned char *skb_pull (struct sk_buff *skb, unsigned int len)
{
    if (len > skb->len)
        return NULL;

    skb->mBufPtr->m_data += len;
    skb->data += len;

    skb->mBufPtr->m_pkthdr.len -= len;
    skb->mBufPtr->m_len -= len;
    skb->len -= len;

    return skb->data;
}

void skb_trim (struct sk_buff *skb, unsigned int len)
{
    if (skb->len > len)
    {
        skb->mBufPtr->m_pkthdr.len = len;
        skb->mBufPtr->m_len = len;
        skb->len = len;

        skb->tail = skb->data + len;
    }
}

struct sk_buff *alloc_skb (unsigned int size, int gfp_mask)
{
    struct sk_buff *skb;                /* newly allocated buffer */
    
    /* Allocate a buffer */
    
    skb = vxskb_alloc ();
    if (skb == NULL)
        return NULL;
    
    /* Allocate an mBlk tuple of the required size */
    
    skb->mBufPtr = vxskb_tuple_get (size);
    if (skb->mBufPtr == NULL)
    {
        vxskb_dealloc (skb);
        return NULL;
    }

    /* Mark buffer as empty */

    skb->data = (unsigned char *)vxskb_head (skb);
    skb->mBufPtr->m_data = (char *)skb->data;

    skb->tail = skb->data;

    skb->mBufPtr->m_pkthdr.len = 0;
    skb->mBufPtr->m_len = 0;
    /* note: skb->len is already 0, so we don't have to set it */

    return skb;
}

void kfree_skb (struct sk_buff *skb)
{
    if (skb != NULL)
    {
		if (skb->mBufPtr != NULL)
        	MBUF_FREE (skb->mBufPtr);
        vxskb_dealloc (skb);
    }
}


int skb_headroom(const struct sk_buff *skb)
{
	return (int)(skb->data - vxskb_head (skb));
}

int skb_tailroom(const struct sk_buff *skb)
{
	return (int)(vxskb_end (skb) - skb->tail);
}

void *skb_header_pointer(const struct sk_buff *skb, int offset,
                                       int len, void *buffer)
{
	assert (skb->mBufPtr->m_ext.ext_size >= (skb_headroom(skb) + offset + len));
	return (skb->data + offset);
}

int pskb_may_pull(struct sk_buff *skb, unsigned int len)
{
	return (skb->len >= len);
}

int pskb_trim(struct sk_buff *skb, unsigned int len)
{
	skb_trim (skb, len);
	return 0;
}

int skb_cloned(const struct sk_buff *skb)
{
	/*
	* netBufLib allows cluster reference count to be read without taking locks
	* (on both UP and SMP systems)
	*/

	return (skb->mBufPtr->m_ext.ext_refcnt);
}

int skb_queue_init(struct socket *sock, int tx_size, int rx_size)
{
	int error = 0;
	if (sock->so_snd.sb_hiwat == 0 || sock->so_rcv.sb_hiwat == 0) 
	{
		sbrelease(&sock->so_rcv, sock);
		sbrelease(&sock->so_snd, sock);

        if((error = sb_max_set(tx_size + rx_size)))
            return error;

	    error = soreserve(sock, tx_size, rx_size);
    }
	return error == 0 ? EOK : error;
}

int skb_queue_init_unused(struct socket *sock, int tx_size, int rx_size)
{
	int error;
	if (sock->so_snd.sb_hiwat == 0 || sock->so_rcv.sb_hiwat == 0) 
	{
		sbrelease(&sock->so_rcv, sock);
		sbrelease(&sock->so_snd, sock);

SKB_QUEUE_INIT:
	    error = soreserve(sock, tx_size, rx_size);
	    if (error)
	    {
			/* set the max socket buffer size higher (TODO: test this feature!!!) */
			int new_max_size;
			if (tx_size > rx_size)
				new_max_size = tx_size;
			else
				new_max_size = rx_size;
				
			if (!sb_max_set (new_max_size))
				return (error);
			
			goto SKB_QUEUE_INIT;	
		}	
	}
	return EOK;
}


uint32_t skb_queue_len(struct sockbuf *sb)
{
	int queue_len;
	struct mbuf *mp = sb->sb_mb;
	for (queue_len=0; mp; queue_len++)	
		mp = mp->m_nextpkt;
	return queue_len;
}


struct sk_buff* __skb_dequeue(struct sockbuf *sb)
{
	struct mbuf *mp;
	struct sk_buff *buf;

	/* copy the first mbuf */
	if (sb->sb_mb == NULL)
		return NULL;

//	sblock(sb, M_WAITOK);
	mp = MBUF_DUP(sb->sb_mb, 0, M_COPYALL, M_DONTWAIT);
	sbdroprecord(sb);
//	sbunlock(sb);

	ASSERT(mp);

	/* associate mbuf with an sk_buff */
	buf = vxskb_attach (mp);
	if (buf && buf->data)
		TIPC_SKB_CB(buf)->handle = msg_data(buf_msg(buf));
	return buf;
}

struct mbuf* skb_peek(struct sockbuf *sb)
{
	struct mbuf *mp;
	ASSERT(sb->sb_cc > 0);

	for (mp=sb->sb_mb; mp; mp = mp->m_next)
	{
		/* look for MT_DATA mbuf */
		if ((mp->m_type == MT_DATA) || (mp->m_type == MT_HEADER))
		{
			/* return ptr to mbuf */
			return mp;
		}	
	}
	return NULL;	
}

int skb_queue_empty(struct sockbuf *sb)
{
	return (sb->sb_cc <= 0);
}

void skb_queue_destroy(struct socket *sock)
{
	sbrelease (&sock->so_rcv, sock);
	sbrelease (&sock->so_snd, sock);
}

